
AnimalDistribution = {};

function AnimalDistribution:InsertSpecialization()
    if g_placeableSpecializationManager:getSpecializationByName("AnimalDistributionSpecialization") == nil then
        g_placeableSpecializationManager:addSpecialization("AnimalDistributionSpecialization", "AnimalDistributionSpecialization", Utils.getFilename("AnimalDistributionSpecialization.lua", g_currentModDirectory), nil)
    end
   
    TypeManager.validateTypes = Utils.appendedFunction(TypeManager.validateTypes, function(self)
        if self.typeName == "placeable" then
            for placeableType, placeable in pairs(g_placeableTypeManager:getTypes()) do
                if SpecializationUtil.hasSpecialization(PlaceableHusbandryAnimals, placeable.specializations) then
                    g_placeableTypeManager:addSpecialization(placeableType, "AnimalDistributionSpecialization")
                end
            end
        end
    end)
end

function AnimalDistribution:loadMap(name)
    addConsoleCommand("AnimalDistributionTest", "AnimalDistributionTest", "AnimalDistributionTest", self)
end

function AnimalDistribution:test()
    print("AnimalDistributionTest")
    g_currentMission.productionChainManager:distributeGoods();
end

function AnimalDistribution:GetAllPlaceableHusbandries() 
    local husbandries = {}
    for i = #g_currentMission.placeableSystem.placeables, 1, -1 do
        local CurrentPlaceable = g_currentMission.placeableSystem.placeables[i]
        if CurrentPlaceable.spec_husbandryAnimals ~= nil and CurrentPlaceable.spec_husbandry ~= nil then
            table.insert(husbandries, CurrentPlaceable)
        end
    end
    return husbandries
end

function AnimalDistribution:GetAllPlaceableHusbandriesInNeed(FarmId,FillTypeId,PlaceableHusbandries)
    local husbandries = {}
    for i = #PlaceableHusbandries, 1, -1 do
        if PlaceableHusbandries[i]:getOwnerFarmId() == FarmId and PlaceableHusbandries[i].spec_husbandryAnimals.FeedFromChain == true then
            if PlaceableHusbandries[i].spec_husbandryFood ~= nil and PlaceableHusbandries[i].spec_husbandryFood:getFreeFoodCapacity(FillTypeId) > 0 then
                table.insert(husbandries, PlaceableHusbandries[i])
            end
        end
    end

    return husbandries
end

function AnimalDistribution:GlobalDistribution()
    if not self.isServer then return end

    local PlaceableHusbandries = AnimalDistribution:GetAllPlaceableHusbandries()
    AnimalDistribution:DistributeFromAnimals(PlaceableHusbandries,self.farmIds);
    AnimalDistribution:DistributeToAnimals(PlaceableHusbandries,self.farmIds);
end

function AnimalDistribution:DivisionDistributionFromAnimals(AmountToDistribute, Reciverse, FillTypeId)
    local DivisionTable = {}
    local TotalTaken = 0
    local TotalFreeCapacity = 0
    local NumReceivers = #Reciverse

    for n = 1, NumReceivers do
        TotalFreeCapacity = TotalFreeCapacity + Reciverse[n].storage:getFreeCapacity(FillTypeId, true)
    end

    if TotalFreeCapacity > 0 and NumReceivers > 0 then
        local EvenDistributionAmount = math.min(AmountToDistribute, TotalFreeCapacity) / NumReceivers

        for n = 1, NumReceivers do
            local prodPointInDemand = Reciverse[n]
            local maxAmountToReceive = prodPointInDemand.storage:getFreeCapacity(FillTypeId, true)

            local amountToTransfer = math.min(maxAmountToReceive, EvenDistributionAmount)
            DivisionTable[n] = amountToTransfer
            TotalTaken = TotalTaken + amountToTransfer
        end
    else
        for n = 1, NumReceivers do
            DivisionTable[n] = 0
        end
    end

    return DivisionTable, TotalTaken
end

function AnimalDistribution:DivisionDistributionToAnimals(AmountToDistribute, Reciverse, FillTypeId)
    local DivisionTable = {}
    local TotalTaken = 0
    local TotalFreeCapacity = 0
    local NumReceivers = #Reciverse

    for n = 1, NumReceivers do
        TotalFreeCapacity = TotalFreeCapacity + Reciverse[n].spec_husbandryFood:getFreeFoodCapacity(FillTypeId)
    end

    if TotalFreeCapacity > 0 and NumReceivers > 0 then
        local EvenDistributionAmount = math.min(AmountToDistribute, TotalFreeCapacity) / NumReceivers

        for n = 1, NumReceivers do
            local maxAmountToReceive = Reciverse[n].spec_husbandryFood:getFreeFoodCapacity(FillTypeId)

            local amountToTransfer = math.min(maxAmountToReceive, EvenDistributionAmount)
            DivisionTable[n] = amountToTransfer
            TotalTaken = TotalTaken + amountToTransfer
        end
    else
        for n = 1, NumReceivers do
            DivisionTable[n] = 0
        end
    end

    return DivisionTable, TotalTaken
end

function AnimalDistribution:DistributeFromAnimals(PlaceableHusbandries, farmIds)
    for FarmId, farmTable in pairs(farmIds) do
        for i = #PlaceableHusbandries, 1, -1 do
            if PlaceableHusbandries[i]:getOwnerFarmId() == FarmId and PlaceableHusbandries[i].spec_husbandryAnimals.FeedToChain == true then
                if PlaceableHusbandries[i].spec_husbandry.storage ~= nil then
                    for _, CurrentFillType in ipairs({FillType.MILK, FillType.LIQUIDMANURE, FillType.MANURE}) do
                        if PlaceableHusbandries[i].spec_husbandry.storage:getIsFillTypeSupported(CurrentFillType) then
                            local AvaibleLevel = PlaceableHusbandries[i].spec_husbandry.storage:getFillLevel(CurrentFillType)
                            local ProdPointsInDemand = farmTable.inputTypeToProductionPoints[CurrentFillType] or {}
                            if AvaibleLevel > 0 and #ProdPointsInDemand > 0 then 
                                local DivisionTable, TotalTaken = AnimalDistribution:DivisionDistributionFromAnimals(AvaibleLevel, ProdPointsInDemand, CurrentFillType);
                                for n = 1, #ProdPointsInDemand do
                                    ProdPointsInDemand[n].storage:setFillLevel(ProdPointsInDemand[n].storage:getFillLevel(CurrentFillType) + DivisionTable[n], CurrentFillType)
                                end
                                PlaceableHusbandries[i].spec_husbandry.storage:setFillLevel(AvaibleLevel-TotalTaken,CurrentFillType) 
                            end
                        end
                    end
                end

                if PlaceableHusbandries[i].spec_husbandryPallets ~= nil then
                    for _, CurrentFillType in ipairs({FillType.EGG, FillType.WOOL}) do 
                        AnimalDistribution.LastPallets = nil
                        PlaceableHusbandries[i].spec_husbandryPallets.palletSpawner:getAllPallets(CurrentFillType, AnimalDistribution.GetAllPalletsCallback, self)              
                        if AnimalDistribution.LastPallets ~= nil and #AnimalDistribution.LastPallets > 0 then
                            for _, Pallet in pairs(AnimalDistribution.LastPallets) do
                                local AvaibleLevel = Pallet:getFillUnitFillLevel(PlaceableHusbandries[i].spec_husbandryPallets.fillUnitIndex)
                                local ProdPointsInDemand = farmTable.inputTypeToProductionPoints[CurrentFillType] or {}
                                if AvaibleLevel > 0 and #ProdPointsInDemand > 0 then 
                                    local DivisionTable, TotalTaken = AnimalDistribution:DivisionDistributionFromAnimals(AvaibleLevel, ProdPointsInDemand, CurrentFillType)
                                    for n = 1, #ProdPointsInDemand do
                                        ProdPointsInDemand[n].storage:setFillLevel(ProdPointsInDemand[n].storage:getFillLevel(CurrentFillType) + DivisionTable[n], CurrentFillType)
                                    end

                                    Pallet:addFillUnitFillLevel(Pallet:getOwnerFarmId(), PlaceableHusbandries[i].spec_husbandryPallets.fillUnitIndex, -TotalTaken, CurrentFillType, ToolType.UNDEFINED)
                                end
                            end
                        end
                    end  
                end
            end
        end
    end
end

function AnimalDistribution:DistributeToAnimals(PlaceableHusbandries, farmIds)
    for FarmId, farmTable in pairs(farmIds) do
		for i = 1, #farmTable.productionPoints do
			local distributingProdPoint = farmTable.productionPoints[i]

			for fillTypeIdToDistribute in pairs(distributingProdPoint.outputFillTypeIdsAutoDeliver) do
				local amountToDistribute = distributingProdPoint.storage:getFillLevel(fillTypeIdToDistribute)

				if amountToDistribute > 0 then
                    local PlaceableHusbandriesInNeed = AnimalDistribution:GetAllPlaceableHusbandriesInNeed(FarmId, fillTypeIdToDistribute, PlaceableHusbandries)               
                    if #PlaceableHusbandriesInNeed > 0 then 
                        local DivisionTable, TotalTaken = AnimalDistribution:DivisionDistributionToAnimals(amountToDistribute, PlaceableHusbandriesInNeed, fillTypeIdToDistribute)
                
                        for n = 1, #PlaceableHusbandriesInNeed do
                            PlaceableHusbandriesInNeed[n]:addFood(PlaceableHusbandriesInNeed[n]:getOwnerFarmId(), DivisionTable[n], fillTypeIdToDistribute, nil,nil,nil )
                        end
                        distributingProdPoint.storage:setFillLevel( distributingProdPoint.storage:getFillLevel(fillTypeIdToDistribute) -TotalTaken, fillTypeIdToDistribute)
                    end
                end
			end
		end
	end
end

function AnimalDistribution:GetAllPalletsCallback(Pallets)
    AnimalDistribution.LastPallets = Pallets
end

ProductionChainManager.distributeGoods = Utils.appendedFunction(ProductionChainManager.distributeGoods, AnimalDistribution.GlobalDistribution)
AnimalDistribution:InsertSpecialization();
addModEventListener(AnimalDistribution);